from typing import Dict, Literal, Optional, Union

import orjson
from pydantic import Json, StrictInt, StrictStr

from mkfst.models.base.message import Message


class HTTPMessage(Message):
    protocol: StrictStr = "HTTP/1.1"
    path: Optional[StrictStr]
    method: Optional[
        Literal["GET", "POST", "HEAD", "OPTIONS", "PUT", "PATCH", "DELETE"]
    ]
    status: Optional[StrictInt]
    status_message: Optional[StrictStr]
    params: Dict[StrictStr, StrictStr] = {}
    headers: Dict[StrictStr, StrictStr] = {}
    data: Optional[Union[Json, StrictStr]]

    def prepare_response(self):
        message = "OK"
        if self.error:
            message = self.error

        head_line = f"HTTP/1.1 {self.status} {message}"

        encoded_data: str = ""

        if isinstance(self.data, Message):
            encoded_data = orjson.dumps(self.data.to_data())

            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif self.data:
            encoded_data = self.data

            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        else:
            headers = "content-length: 0"

        response_headers = self.headers
        if response_headers:
            for key in response_headers:
                headers = f"{headers}\r\n{key}: {response_headers[key]}"

        return f"{head_line}\r\n{headers}\r\n\r\n{encoded_data}".encode()



import time
from typing import Any, Optional

import orjson
from pydantic import (
    BaseModel,
    StrictFloat,
    StrictInt,
    StrictStr,
)

from .run_status import RunStatus


class TaskRun(BaseModel):
    run_id: StrictInt
    status: RunStatus
    error: Optional[StrictStr]=None
    trace: Optional[StrictStr]=None
    start: StrictInt | StrictFloat = time.monotonic()
    end: Optional[StrictInt | StrictFloat] = None
    elapsed: StrictInt | StrictFloat = 0
    result: Optional[Any] = None

    def to_json(self):

        return orjson.dumps({
            'run_id': self.run_id,
            'status': self.status.value,
            'error': self.error,
            'start': self.start,
            'end': self.end,
            'elapsed': self.elapsed
        })
    
    def to_data(self):
        return {
            'run_id': self.run_id,
            'status': self.status.value,
            'error': self.error,
            'start': self.start,
            'end': self.end,
            'elapsed': self.elapsed
        }


from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)

from pydantic import BaseModel

from .base_wrapper import BaseWrapper
from .response_context import ResponseContext
from .types import Handler, MiddlewareHandler, MiddlewareType

T = TypeVar("T")


class UnidirectionalWrapper(BaseWrapper):
    def __init__(
        self,
        name: str,
        handler: Handler,
        middleware_type: MiddlewareType = MiddlewareType.UNIDIRECTIONAL_BEFORE,
        methods: Optional[
            List[
                Literal[
                    "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
                ]
            ]
        ] = None,
        responses: Optional[Dict[int, BaseModel]] = None,
        serializers: Optional[Dict[int, Callable[..., str]]] = None,
        response_headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__()

        self.name = name
        self.path = handler.path
        self.methods: List[
            Literal["GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"]
        ] = handler.methods

        if methods:
            self.methods.extend(methods)

        self.response_headers: Union[Dict[str, str], None] = handler.response_headers

        if self.response_headers and response_headers:
            self.response_headers.update(response_headers)

        elif response_headers:
            self.response_headers = response_headers

        self.responses = responses
        self.serializers = serializers
        self.limit = handler.limit

        self.handler = handler
        self.wraps = isinstance(handler, BaseWrapper)

        if self.handler.response_headers and self.response_headers:
            self.handler.response_headers = {}

        self.run: MiddlewareHandler | None = None
        self.middleware_type = middleware_type

    async def __call__(
        self, 
        context: ResponseContext | None = None,
        response: Any | None = None,
    ):
        if context is None:
            raise Exception('Err. - Context is missing.')
        
        if self.wraps and self.middleware_type == MiddlewareType.UNIDIRECTIONAL_AFTER:
            # Is wraps additional middleware so expect
            # a middleware response.
            (
                context, 
                response
            ) = await self.handler(
                context=context,
                response=response,
            )


            (
                context,
                response
            ), _ = await self.run(
                context=context,
                response=response,
                handler=self.handler,
            )

            return (
                context, 
                response,
            )

        elif self.wraps and self.middleware_type == MiddlewareType.UNIDIRECTIONAL_BEFORE:
            # Wraps a call and should be executed before.
            
            (
                context, 
                response
            ), run_next = await self.run(
                context=context,
                response=response,
                handler=self.handler,
            )

            if run_next is False:
                return context, response

            (
                context,
                response
            ) = await self.handler(
                context=context,
                response=response,
            )

            return (
                context, 
                response,
            )

        elif self.middleware_type == MiddlewareType.UNIDIRECTIONAL_AFTER:
            # Wraps a call and should be executed after.

            response = await self.handler(
                *context.args,
                **context.kwargs
            )

            (
                context, 
                response
            ), _ = await self.run(
                context=context,
                response=response,
                handler=self.handler,
            )

            return (
                context, 
                response,
            )
        
        else:
            (
                context, 
                response
            ), run_next = await self.run(
                context=context,
                response=response,
                handler=self.handler,
            )

            response = await self.handler(
                *context.args,
                **context.kwargs
            )

            return (
                context, 
                response,
            )



import functools
from collections import defaultdict
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    TypeVar,
    get_args,
    get_type_hints,
)

from pydantic import BaseModel

from mkfst.models.http import Limit

from .get_content_type import get_content_type

T = TypeVar("T")


def endpoint(
    path: Optional[str] = "/",
    methods: List[
        Literal[
            "GET", 
            "HEAD", 
            "OPTIONS", 
            "POST", "PUT", 
            "PATCH", 
            "DELETE", 
            "TRACE",
        ],
    ] = ["GET"],
    responses: Optional[Dict[int, BaseModel]] = None,
    response_headers: Optional[Dict[str, str]] = None,
    limit: Optional[Limit] = None,
    summary: str | None = None,
    tags: List[str] | None = None,
    method_metadata: Dict[
        Literal[
            "GET", 
            "HEAD", 
            "OPTIONS", 
            "POST", "PUT", 
            "PATCH", 
            "DELETE", 
            "TRACE",
        ],
        Dict[
            Literal[
                'name',
                'tags',
                'description',
                'summary',
                'docs_url',
                'docs_description',
                'depreciated',
            ],
            str | List[str] | bool
        ]
    ] | None = None,
    additional_docs: Dict[
        Literal[
            'doc_description',
            'docs_url'
        ],
        str
    ] | None = None,
    depreciated: bool = False
):

    def wraps(func):

        return_type = get_type_hints(func).get('return')
        return_args = get_args(return_type)

        if isinstance(return_args, tuple) and len(return_args) > 0:
            return_type = return_args[0]

        headers = response_headers or {}

        lowered_headers: Dict[str, str] = {}
        lowered_headers = {
            key.lower(): value.lower() for key, value in headers.items()
        }

        if lowered_headers.get('content-type') is None:
            headers.update({
                'content-type': get_content_type(return_type)
            })
            
        func.as_endpoint = True
        func.path = path
        func.methods = methods
        
        func.response_headers = headers
        func.responses = responses
        func.limit = limit
        func.summary = summary
        func.tags = tags
        func.method_metadata = method_metadata or defaultdict(dict)
        func.additional_docs = additional_docs,
        func.depreciated = depreciated


        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wraps



import asyncio
import functools
from typing import Literal

from pydantic import BaseModel, StrictBytes, StrictStr


class FileUpload(BaseModel):
    data: StrictBytes | StrictStr
    content_type: StrictStr | None = None
    encoding: StrictStr | None = None

    async def upload(
        self, 
        path: str, 
        read_type: Literal['string', 'binary']='string',
    ):
        upload_file = await asyncio.to_thread(
            functools.partial(
                self._upload,
                path,
                read_type
            )
        )

        return FileUpload(
            data=upload_file
        )

    def _upload( 
        self, 
        path: str, 
        read_type: Literal['string', 'binary']='string',
    ):
        
        read_mode = 'r'
        if read_type == 'binary':
            read_mode = 'rb'

        with open(path, read_mode) as upload_file:
            return upload_file.read()
         


from __future__ import annotations

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Tuple,
    Type,
)

from mkfst.connection.tcp.fabricator import Fabricator
from mkfst.models.http.request_models import Cookies, Headers


class ResponseContext:
    
    __slots__ = (
        'path',
        'method',
        'params',
        'query',
        'request_headers',
        'cookies',
        'body',
        'response_headers',
        'status',
        'errors',
        'args',
        'kwargs',
        '_data',
        'fabricator',
        'parser',
        'compressor',
        'compression_level',
        'ip_address',
    )

    def __init__(
        self,
        path: str,
        method: str,
        params: Dict[str, str] | None,
        query: str | None,
        data: List[bytes],
        args: Tuple[Any, ...], 
        kwargs: Dict[str, Any],
        fabricator: Fabricator,
        parser: Type[Any],
        ip_address: str,
    ) -> None:
        self.ip_address = ip_address
        self.path = path
        self.method = method
        self.params: Dict[str, Any] = params or {}
        self.query: str = query or ''

        self.request_headers: Dict[str, Any] = {}
        self.cookies: Dict[str, Any] = {}
        self.body: Any | None = None

        self.response_headers: Dict[str, Any] = {}
        self.status: int | None = None
        self.parser = parser
        self.errors: List[Exception] = []

        self.args = args
        self.kwargs = kwargs
        self._data = data
        self.fabricator = fabricator
        self.compressor: Literal['gzip', 'zstd'] | None = None
        self.compression_level: int | None = None

    def update(self, context: ResponseContext):
        self.request_headers.update(context.request_headers)
        self.response_headers.update(context.response_headers)
        self.params.update(context.params)
        self.cookies.update(context.cookies)

        if context.status:
            self.status = context.status

    def get_bytes_arg(self) -> Tuple[bytes, int, None] | Tuple[bytes, None, int] | Tuple[bytes, None, None]:
        for arg in self.args:
            if isinstance(arg, bytes):
                return arg
            
        for value in self.kwargs.values():
            if isinstance(value, bytes):
                return value
            
        return b''
    
    def get_headers(self):
        if (
            param_key := self.fabricator.param_keys.get('headers')
        ):
            
            is_position = isinstance(param_key, int)
            headers: Headers = (
                self.args[param_key]
                if is_position
                else self.kwargs[param_key]
            )
            
            return headers
            
            
        return None
    
    def get_headers_and_cookies(self):

        headers: Dict[str, Any] = {}
        cookies: Dict[str, Any] = {}

        if (
            param_key := self.fabricator.param_keys.get('headers')
        ):
            headers: Headers = (
                self.args[param_key]
                if isinstance(param_key, int)
                else self.kwargs[param_key]
            )

        elif (
            param_key := self.fabricator.param_keys.get('cookies')
        ):
            cookies: Cookies = (
                self.args[param_key]
                if isinstance(param_key, int)
                else self.kwargs[param_key]
            )

        headers_content: Dict[str, str] = {}
        cookies_content: Dict[str, str] = {}
        if headers:
            headers_content = headers.model_dump()

        if headers and cookies is None:
            cookies_content = Cookies.make_raw(headers)
        
        elif cookies:
            cookies_content = cookies.model_dump()
            
        return (
            headers_content,
            cookies_content,
        )
    
    def update_request_headers(
        self,
        new_headers: Dict[str, Any]
    ):
        
        self.request_headers.update(new_headers)

        headers = self.get_headers()

        updated_headers: Headers | None = None

        if headers:
            headers_dict = headers.model_dump()

            headers_dict.update(self.request_headers)

            updated_headers = headers.model_copy(headers_dict)

        param_key = self.fabricator.param_keys.get('headers')
        is_position = isinstance(param_key, int)

        if updated_headers and param_key and is_position:
            self.args[param_key] = updated_headers

        elif updated_headers and param_key:
            self.kwargs[param_key] = updated_headers

    def update_request_data(
        self,
        data: Any
    ):
        body_key = self.fabricator.param_keys.get('body')
        is_position = isinstance(body_key, int)
        
        if body_key and is_position:
            self.args[body_key] = data

        elif body_key:
            self.kwargs[body_key] = data

            


from typing import Any, Dict, List, Literal

from pydantic import BaseModel, StrictBytes, StrictInt, StrictStr

from mkfst.models import (
        HTML,
        Body,
        FileUpload,
        Headers,
        Parameters,
        Query,
)

from .parsed_endpoint_metadata import (
        HTTPMethod,
        ParsedEndpointMetadata,
)


class ParsedEndpoint(BaseModel):
    path: StrictStr
    methods: List[HTTPMethod]
    endpoint_metadata: ParsedEndpointMetadata
    responses: Dict[
        StrictInt, 
        type[HTML]
        | type[FileUpload]
        | type[Body]
        | type[BaseModel]
        | type[dict]
        | type[list]
        | type[str]
        | type[bytes]
        | StrictStr
        | StrictBytes
        | None
    ]
    response_headers: Dict[StrictStr, Any] | None = None
    headers: type[Headers] | None = None
    parameters: type[Parameters] | None = None
    query: type[Query] | None = None
    body: (
        type[HTML]
        | type[FileUpload]
        | type[Body]
        | type[BaseModel]
        | type[dict]
        | type[list]
        | type[str]
        | type[bytes]
        | StrictStr
        | StrictBytes
        | None
    ) = None
    required: List[
        Literal[
            "parameters",
            "headers",
            "query",
            "body",
        ]
    ] | None = None

    def to_dict(self):
        return {
            field: getattr(self, field) for field in self.model_fields
        }


import asyncio

from mkfst import (
    Service,
    endpoint,
    Env,
)
from pydantic import BaseModel


class Greeting(BaseModel):
    message: str


class TestService(Service):
    @endpoint("/")
    async def get_service(self) -> str:
        return "Hello World"

    @endpoint("/post", methods=["POST"])
    async def get_data(self, data: dict) -> dict:
        return data


if __name__ == "__main__":

    async def run_server():
        server = TestService(
            "localhost",
            6099,
            log_level="error",
            env=Env(
                MERCURY_SYNC_ENABLE_REQUEST_CACHING=True,
            ),
        )

        await server.run()

    asyncio.run(run_server())



import os
from pydantic import BaseModel
from typing import Dict, Type, TypeVar, Union

from dotenv import dotenv_values

from .env import Env

T = TypeVar("T", bound=BaseModel)

PrimaryType = Union[str, int, bool, float, bytes]


def load_env(env: Type[T], env_file: str = None, existing: Env | None = None) -> T:
    env_type: Env = env
    envars = env_type.types_map()

    if env_file is None:
        env_file = ".env"

    values: Dict[str, PrimaryType] = {}
    for envar_name, envar_type in envars.items():
        envar_value = os.getenv(envar_name)
        if envar_value:
            values[envar_name] = envar_type(envar_value)

    if env_file and os.path.exists(env_file):
        env_file_values = dotenv_values(dotenv_path=env_file)

        for envar_name, envar_value in env_file_values.items():
            envar_type = envars.get(envar_name)
            if envar_type:
                env_file_values[envar_name] = envar_type(envar_value)

        values.update(env_file_values)

    if existing:
        values.update(**existing.model_dump())

    return env(**{name: value for name, value in values.items() if value is not None})



from pydantic import BaseModel, RootModel, StrictStr, conlist


class InternalError(BaseModel):
    error: StrictStr

class InternalErrorSet(RootModel):
    root: conlist(InternalError, min_length=1)
    


from typing import List, Set

from mkfst.models.http import Headers, Parameters, Query

from .models import Parameter
from .parsed_types import FieldMetadata, FieldsMetadata, FieldType, PropertyMetadata


def parse_type(field: FieldMetadata) -> List[str] | str:
    field_data: List[FieldType] = field.get("anyOf")
    if field_data:
        return {
            'enum': [
                field_type.get("type") for field_type in field_data
            ]
        }
    
    return {
        'type': field.get("type", "string")
    }


def format_header_name(name: str):
    return '-'.join([
        segment.capitalize() for segment in name.split('_')
    ])

def parse_param(
    param: type[Headers] | type[Parameters] | type[Query],
    path_params: Set[str]
) -> Parameter:

    schema: PropertyMetadata = param.model_json_schema()
    metadata: FieldsMetadata = schema.get('properties')
    required = schema.get('required', [])

    fields = list(param.model_fields.keys())

    if param == Parameters:
        fields = [
            field for field in fields if field in path_params
        ]

        additional_params = list(filter(
            lambda path_param: path_param not in fields,
            path_params
        ))

        for path_param in additional_params:
            metadata[path_param] = {
                "type": "string",
                "format": "string",
            }

            fields.append(path_param)

    location: str | None = None

    if param == Headers or param in Headers.__subclasses__():
        location = 'header'

    elif param == Query or param in Query.__subclasses__():
        location = 'query'

    elif param == Parameters or param in Parameters.__subclasses__():
        location = 'path'

    else:
        location = 'cookie'

    return [
        {
            'name': format_header_name(
                field
            ) if location == 'header' else field,
            'description': f'{field} header',
            'required': field in required,
            'in': location,
            'schema': {
                **parse_type(
                    metadata.get(field)
                ),
                'format': metadata.get(
                    field, {}
                ).get('format')
            }
        } for field in fields
    ]


import os
from typing import Callable, Dict, Literal, Union

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    IPvAnyAddress,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)

PrimaryType = Union[str, int, float, bytes, bool]


class Env(BaseModel):
    MERCURY_SYNC_SERVER_URL: AnyHttpUrl | None = None
    MERCURY_SYNC_API_VERISON: StrictStr = "0.0.1"
    MERCURY_SYNC_TASK_EXECUTOR_TYPE: Literal["thread", "process", "none"] = "process"
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY: StrictFloat = 2
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW: StrictStr = "1m"
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD: Union[
        StrictInt, StrictFloat
    ] = 0.2
    MERCURY_SYNC_HTTP_HANDLER_TIMEOUT: StrictStr = "1m"
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal[
        "none", "global", "endpoint", "ip", "ip-endpoint", "custom"
    ] = "none"
    MERCURY_SYNC_HTTP_RATE_LIMITER_TYPE: Literal[
        "adaptive",
        "cpu-adaptive",
        "leaky-bucket",
        "rate-adaptive",
        "sliding-window",
        "token-bucket",
    ] = "sliding-window"
    MERCURY_SYNC_HTTP_CORS_ENABLED: StrictBool = False
    MERCURY_SYNC_HTTP_MEMORY_LIMIT: StrictStr = "512mb"
    MERCURY_SYNC_HTTP_CPU_LIMIT: Union[StrictFloat, StrictInt] = 50
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE: StrictInt = 10
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF: StrictStr = "1s"
    MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD: StrictStr = "1s"
    MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS: StrictInt = 100
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal[
        "ip",
        "endpoint",
        "global",
        "ip-endpoint",
        "none",
    ] = "none"
    MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT: StrictBool = True
    MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION: StrictBool = False
    MERCURY_SYNC_USE_HTTP_SERVER: StrictBool = True
    MERCURY_SYNC_USE_HTTP_AND_TCP_SERVERS: StrictBool = False
    MERCURY_SYNC_USE_UDP_MULTICAST: StrictBool = False
    MERCURY_SYNC_TCP_CONNECT_RETRIES: StrictInt = 3
    MERCURY_SYNC_CLEANUP_INTERVAL: StrictStr = "0.5s"
    MERCURY_SYNC_MAX_CONCURRENCY: StrictInt = 2048
    MERCURY_SYNC_AUTH_SECRET: StrictStr = "testtoken"
    MERCURY_SYNC_MULTICAST_GROUP: IPvAnyAddress = "224.1.1.1"
    MERCURY_SYNC_LOGS_DIRECTORY: StrictStr = os.getcwd()
    MERCURY_SYNC_REQUEST_TIMEOUT: StrictStr = "30s"
    MERCURY_SYNC_LOG_LEVEL: StrictStr = "info"
    MERCURY_SYNC_TASK_RUNNER_MAX_THREADS: StrictInt = os.cpu_count()
    MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE: StrictInt = 100
    MERCURY_SYNC_ENABLE_REQUEST_CACHING: StrictBool = False

    @classmethod
    def types_map(cls) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            "MERCURY_SYNC_SERVER_URL": str,
            "MERCURY_SYNC_API_VERISON": str,
            "MERCURY_SYNC_TASK_EXECUTOR_TYPE": str,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY": float,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW": str,
            "MERCURY_SYNC_HTTP_HANDLER_TIMEOUT": str,
            "MERCURY_SYNC_USE_UDP_MULTICAST": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD": float,
            "MERCURY_SYNC_HTTP_CORS_ENABLED": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_HTTP_MEMORY_LIMIT": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE": int,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF": str,
            "MERCURY_SYNC_HTTP_CPU_LIMIT": float,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD": str,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS": int,
            "MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_USE_HTTP_SERVER": lambda value: True
            if value.lower() == "true"
            else False,
            "MERCURY_SYNC_TCP_CONNECT_RETRIES": int,
            "MERCURY_SYNC_CLEANUP_INTERVAL": str,
            "MERCURY_SYNC_MAX_CONCURRENCY": int,
            "MERCURY_SYNC_AUTH_SECRET": str,
            "MERCURY_SYNC_MULTICAST_GROUP": str,
            "MERCURY_SYNC_LOGS_DIRECTORY": str,
            "MERCURY_SYNC_REQUEST_TIMEOUT": str,
            "MERCURY_SYNC_LOG_LEVEL": str,
            "MERCURY_SYNC_TASK_RUNNER_MAX_THREADS": int,
            "MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE": int,
            "MERCURY_SYNC_ENABLE_REQUEST_CACHING": lambda value: True
            if value.lower() == "true"
            else False,
        }



from typing import Any, Type

from pydantic import BaseModel

from mkfst.models import HTML, FileUpload


def get_content_type(return_type: Type[Any]):
    if return_type == FileUpload or return_type in FileUpload.__subclasses__():
        return 'application/octet-stream'

    elif return_type == HTML or return_type in HTML.__subclasses__():
        return 'text/html'

    elif return_type in BaseModel.__subclasses__():
        return 'application/json'
    
    elif return_type == str or isinstance(return_type, str):
        return 'text/plain'
    
    elif return_type == bytes or isinstance(return_type, bytes):
        return 'application/octet-stream'
    
    else:
        return 'application/json'



from enum import Enum
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse

import orjson
from pydantic import AnyHttpUrl

from mkfst.models.base.message import Message

from .http_message import HTTPMessage


class HTTPRequestMethod(Enum):
    GET = "GET"
    POST = "POST"


class HTTPRequest(Message):
    url: AnyHttpUrl
    method: HTTPRequestMethod
    params: Optional[Dict[str, str]]
    headers: Dict[str, str] = {}
    data: Optional[Union[str, Message]]

    class Config:
        arbitrary_types_allowed = True

    def prepare_request(self):
        parsed = urlparse(self.url)

        path = parsed.path
        if path is None:
            path = "/"

        if self.params:
            params_string = "&".join([f"{name}={value}" for name, value in self.params])

            path = f"{path}?{params_string}"

        request: List[str] = [f"{self.method.value} {path} HTTP/1.1"]

        request.append(f"host: {parsed.hostname}")

        request.extend([f"{key}: {value}" for key, value in self.headers.items()])

        encoded_data = None
        if isinstance(self.data, Message):
            encoded_data = orjson.dumps(self.data.to_data())

            request.append("content-type: application/msync")

        elif self.data:
            encoded_data = self.data
            content_length = len(encoded_data)

            request.append(f"content-length: {content_length}")

        request.append("\r\n")

        if encoded_data:
            request.append(encoded_data)

        encoded_request = "\r\n".join(request)

        return encoded_request.encode()

    @classmethod
    def parse(cls, data: bytes):
        response = data.split(b"\r\n")

        response_line = response[0]

        headers: Dict[bytes, bytes] = {}

        header_lines = response[1:]
        data_line_idx = 0

        for header_line in header_lines:
            if header_line == b"":
                data_line_idx += 1
                break

            key, value = header_line.decode().split(":", maxsplit=1)
            headers[key.lower()] = value.strip()

            data_line_idx += 1

        data = b"".join(response[data_line_idx + 1 :]).strip()

        request_type, status, message = response_line.decode().split(" ")

        return HTTPMessage(
            protocol=request_type,
            status=int(status),
            status_message=message,
            headers=headers,
            data=data.decode(),
        )

    @classmethod
    def parse_request(cls, data: bytes):
        response = data.split(b"\r\n")

        response_line = response[0]

        headers: Dict[bytes, bytes] = {}

        header_lines = response[1:]
        data_line_idx = 0

        for header_line in header_lines:
            if header_line == b"":
                data_line_idx += 1
                break

            key, value = header_line.decode().split(":", maxsplit=1)
            headers[key.lower()] = value.strip()

            data_line_idx += 1

        data = b"".join(response[data_line_idx + 1 :]).strip()

        method, path, request_type = response_line.decode().split(" ")

        if path is None or path == "":
            path = "/"

        return HTTPMessage(
            method=method,
            path=path,
            protocol=request_type,
            headers=headers,
            data=data.decode(),
        )



from typing import Callable, List, Literal, Optional, Union

from pydantic import (
    BaseModel,
    IPvAnyAddress,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)

from mkfst.env.memory_parser import MemoryParser
from mkfst.env.time_parser import TimeParser

HTTPMethod = Literal[
    "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
]


class Limit(BaseModel):
    max_requests: Optional[StrictInt] = 1000
    min_requests: Optional[StrictInt] = 100
    request_period: StrictStr = "1s"
    reject_requests: StrictBool = True
    request_backoff: StrictStr = "1s"
    cpu_limit: Optional[Union[StrictFloat, StrictInt]] = None
    memory_limit: Optional[StrictStr] = None
    limiter_type: Literal[
        "adaptive",
        "cpu-adaptive",
        "leaky-bucket",
        "rate-adaptive",
        "sliding-window",
        "token-bucket",
    ] = "sliding-window"
    limit_key: Optional[
        Callable[
            [
                str,
                str,
                IPvAnyAddress,
            ],
            str,
        ]
    ] = None
    rules: Optional[
        List[
            Callable[
                [
                    str,
                    str,
                    IPvAnyAddress,
                ],
                bool,
            ]
        ]
    ] = None

    @property
    def backoff(self):
        return TimeParser(self.request_backoff).time

    @property
    def period(self):
        return TimeParser(self.request_period).time

    @property
    def memory(self):
        return MemoryParser(self.memory_limit).megabytes(accuracy=4)

    def get_key(
        self, 
        path: str,
        method: str, 
        ip_address: IPvAnyAddress, 
        default: str = "default"
    ):
        if self.limit_key is None:
            return default

        return self.limit_key(path, method, ip_address)

    def matches(
        self,
        path,
        method, 
        ip_address: IPvAnyAddress
    ):
        if self.rules is None:
            return True

        matches_rules = False

        for rule in self.rules:
            matches_rules = rule(
                path,
                method, 
                ip_address,
            )

        return matches_rules



from __future__ import annotations

from http.cookies import SimpleCookie
from typing import (
    Any,
    Dict,
    List,
    Type,
)

import orjson
from pydantic import (
    BaseModel,
    StrictBool,
    StrictBytes,
    StrictFloat,
    StrictInt,
    StrictStr,
    model_validator,
)

HTTPEncodable = StrictStr | StrictInt | StrictBool | StrictFloat | None


class Headers(BaseModel):
    @model_validator(mode="before")
    def validate_headers(cls, val: Type[Headers]):
        encodable_values = [
            str,
            int,
            bytes,
            bool,
            float,
            StrictStr,
            StrictInt,
            StrictBool,
            StrictFloat,
        ]

        fields = val.model_fields
        for field, value in fields.items():
            assert value.annotation in encodable_values, (
                f"Err. - field {field} must have JSON encodable type."
            )

    @classmethod
    def make(
        cls,
        model: Type[Headers],
        data: dict[bytes, bytes],
        data_line_idx: int,
    ):
        headers: Dict[str, Any] = {}
        if data_line_idx == -1:
            header_lines = data[1:]
            data_line_idx = 0

            for header_line in header_lines:
                if header_line == b"":
                    data_line_idx += 1
                    break

                key, value = header_line.decode().split(":", maxsplit=1)

                headers[key.lower().replace("-", "_")] = value.strip()

                data_line_idx += 1

            data_line_idx += 1

            return (
                model(**headers),
                data_line_idx,
                headers,
            )

        return (None, data_line_idx, headers)

    @classmethod
    def make_raw(
        cls,
        data: List[bytes],
        data_line_idx: int,
    ):
        headers: Dict[str, Any] = {}
        if data_line_idx == -1:
            header_lines = data[1:]
            data_line_idx = 0

            for header_line in header_lines:
                if header_line == b"":
                    data_line_idx += 1
                    break

                key, value = header_line.decode().split(":", maxsplit=1)

                headers[key.lower().replace("-", "_")] = value.strip()

                data_line_idx += 1

            data_line_idx += 1

            return (
                data_line_idx,
                headers,
            )

        return (data_line_idx, headers)


class Cookies(BaseModel):
    @model_validator(mode="before")
    def validate_headers(cls, val: Type[Query]):
        encodable_values = [
            str,
            int,
            bytes,
            bool,
            float,
            StrictStr,
            StrictInt,
            StrictBool,
            StrictFloat,
        ]

        fields = val.model_fields
        for field, value in fields.items():
            if field == "cookie":
                assert value.annotation == str, (
                    "Err. - field cookie must be either str or StrictStr type"
                )

            else:
                assert value.annotation in encodable_values, (
                    f"Err. - field {field} must have JSON encodable type."
                )

    @classmethod
    def make(cls, model: Type[Cookies], headers: Dict[bytes, bytes]):
        cookie_header = headers.get(b"cookie")

        if cookie_header is None:
            return model()

        parsed_cookies = SimpleCookie()
        parsed_cookies.load(cookie_header)

        return model(**{name: morsel.value for name, morsel in parsed_cookies.items()})

    @classmethod
    def make_raw(cls, headers: Dict[str, Any]):
        cookie_header: str = headers.get("cookie")

        if cookie_header is None:
            return

        parsed_cookies = SimpleCookie()
        parsed_cookies.load(cookie_header)

        return {name: morsel.value for name, morsel in parsed_cookies.items()}


class Parameters(BaseModel):
    @model_validator(mode="before")
    def validate_headers(cls, val: Type[Parameters]):
        encodable_values = [
            str,
            int,
            bytes,
            bool,
            float,
            StrictStr,
            StrictInt,
            StrictBool,
            StrictFloat,
        ]

        fields = val.model_fields
        for field, value in fields.items():
            assert value.annotation in encodable_values, (
                f"Err. - field {field} must have JSON encodable type."
            )


class Query(BaseModel):
    @model_validator(mode="before")
    def validate_headers(cls, val: Type[Query]):
        encodable_values = [
            str,
            int,
            bytes,
            bool,
            float,
            StrictStr,
            StrictInt,
            StrictBool,
            StrictFloat,
        ]

        fields = val.model_fields
        for field, value in fields.items():
            assert value.annotation in encodable_values, (
                f"Err. - field {field} must have JSON encodable type."
            )

    @classmethod
    def make(cls, model: Type[Query], query: str):
        query_params: Dict[str, str] = {}
        if len(query) < 1:
            params = query.split("&")

            for param in params:
                key, value = param.split("=")

                query_params[key] = value

        return model(**query_params)


class Body(BaseModel):
    content: StrictStr | StrictBytes

    def text(self, encoding: str = "utf-8"):
        if isinstance(self.content, bytes):
            return self.content.decode(encoding=encoding)

        return self.content

    def json(self):
        content = self.content
        if isinstance(content, bytes):
            content = self.content.decode()

        return orjson.loads(content)

    @classmethod
    def make(
        cls,
        data: List[bytes],
        data_line_idx: int,
    ):
        return Body(content=b"".join(data[data_line_idx:]).strip())

    @classmethod
    def read(
        cls,
        data: List[bytes],
        data_line_idx: int,
    ):
        if data_line_idx == -1:
            header_lines = data[1:]
            data_line_idx = 0

            headers: Dict[str, Any] = {}

            for header_line in header_lines:
                if header_line == b"":
                    data_line_idx += 1
                    break

                key, value = header_line.decode().split(":", maxsplit=1)

                headers[key.lower()] = value.strip()

                data_line_idx += 1

            data_line_idx += 1

        return b"".join(data[data_line_idx:]).strip()



from typing import Any, Dict, Literal

from mkfst.logging import Entry, LogLevel

JSONEncodableValue = str | int | bool | float | None

class Event(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO


class Request(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO
    path: str | None = None
    protocol: Literal['HTTP/1.1']='HTTP/1.1'
    method: str | None = None
    headers: Dict[str, JSONEncodableValue] = None
    params: Dict[str, JSONEncodableValue] = None
    query: str | None = None
    body: str | bytes | None = None
    error: Exception | str | None = None
    ip_address: str | None = None

    def to_template(
        self,
        template: str,
        context: Dict[str, Any] | None = None,
    ):

        kwargs = {
            'level': self.level.value,
            'path': self.path,
            'method': self.method,
            'headers': ', '.join([
                f'{key}:{header}' for key, header in self.headers.items()
            ]),
            'params': ', '.join([
                f'{key}:{header}' for key, header in self.headers.items()
            ]),
            'query': self.query,
            'body': self.body,
            'error': self.error,
            'ip_address': self.ip_address,
        }

        if context:
            kwargs.update(context)

        return template.format(**kwargs)



class Response(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO
    path: str | None = None
    protocol: Literal['HTTP/1.1']='HTTP/1.1'
    method: str | None = None
    error: Exception | str | None = None
    ip_address: str | None = None
    status: int | None = None

    def to_template(
        self,
        template: str,
        context: Dict[str, Any] | None = None,
    ):
        kwargs: Dict[
            str,
            int | str | bool | float | LogLevel | list | dict | set | Any,
        ] = {field: getattr(self, field) for field in self.__struct_fields__ if getattr(self, field) is not None}

        kwargs["level"] = kwargs["level"].value

        if context:
            kwargs.update(context)

        return template.format(**kwargs)




import inspect
from base64 import b64encode
from gzip import compress as gzip_compress
from typing import Dict, Literal, Optional, Union

import orjson
from pydantic import BaseModel, Json, StrictInt, StrictStr
from zstandard import compress as zstd_compress

from mkfst.models.base.message import Message


class HTTPResponse(BaseModel):
    protocol: StrictStr = "HTTP/1.1"
    path: Optional[StrictStr] = None
    error: Optional[StrictStr] = None
    method: Optional[
        Literal["GET", "POST", "HEAD", "OPTIONS", "PUT", "PATCH", "DELETE"]
    ]=None
    status: Optional[StrictInt] = None
    status_message: Optional[StrictStr] = None
    params: Dict[StrictStr, StrictStr] = {}
    headers: Dict[StrictStr, StrictStr] = {}
    data: Optional[Union[Json, StrictStr]] = None

    def prepare_response(
        self,
        compression: Literal['gzip', 'zstd'] | None = None,
        compression_level: int | None = None
    ):
        message = "OK"
        if self.error:
            message = self.error

        head_line = f"HTTP/1.1 {self.status} {message}"

        encoded_data: str = ""

        if isinstance(self.data, Message):
            encoded_data = orjson.dumps(self.data.to_data()).decode()

            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif inspect.isclass(self.data) and issubclass(self.data, BaseModel):
            encoded_data = orjson.dumps(self.data.model_dump()).decode()
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif isinstance(self.data, (dict, list)):
            encoded_data = orjson.dumps(self.data).decode()
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif self.data:
            encoded_data = self.data

            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        else:
            headers = "content-length: 0"

        if compression == 'gzip':
            encoded_data = b64encode(
                gzip_compress(
                    encoded_data.encode(), 
                    compresslevel=compression_level
                )
            ).decode()
            
            headers = f"{headers}\r\nx-compression-encoding: {compression}"
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif compression == 'zstd':
            encoded_data = b64encode(
                zstd_compress(
                    encoded_data.encode(),
                    level=compression_level
                )
            ).decode()

            headers = f"{headers}\r\nx-compression-encoding: {compression}"
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"


        response_headers = self.headers
        if response_headers:
            for key in response_headers:
                headers = f"{headers}\r\n{key}: {response_headers[key]}"

        return f"{head_line}\r\n{headers}\r\n\r\n{encoded_data}".encode()



from typing import Any, Dict, Type

import orjson
from pydantic import BaseModel

from .file_upload import FileUpload
from .html import HTML


def parse_response(
    response: BaseModel | Dict[Any, Any] | str, 
    response_model: Type[BaseModel | FileUpload | HTML | dict | list | str | bytes ],
) -> str:

    if response_model == HTML or isinstance(response, HTML):
        return response.format()
    
    elif (
        response_model == FileUpload or isinstance(response, FileUpload) 
    ) and isinstance(response.data, bytes):
        return response.data.decode(response.encoding)
    
    elif response_model in FileUpload.__subclasses__():
        return response.data
    
    elif response_model == dict or response_model == list:
        return orjson.dumps(response).decode()
    
    if response_model in BaseModel.__subclasses__():
        return orjson.dumps(response.model_dump()).decode()


    return response
